from rest_framework import status
from rest_framework_filters.filters import RelatedFilter
from django.http import JsonResponse
from django.conf import settings


class InvalidLookupMixin:
    """ A mixin for Django Rest Framework viewsets to check query parameters and return an error if any query parameter
    is not a included in defined in a filter_class (typically defined in your filterset), and element of
    filter_fields (typically set in your viewset), or a valid model field.
    Order of precedence is: filter_class, filter_fields, model field.

    example usage:
        class MyModelViewSet(InvalidLookupMixin, viewsets.ReadOnlyModelViewSet):
    """
    request = None
    model = None
    queryset = None
    filter_class = None
    filter_fields = []

    def get_lookup_expression(self, fs_filter, related_field=None, lookup_expression_list=None):
        """ get lookup expressions as defined in a FilterSet filter """
        if not lookup_expression_list:
            lookup_expression_list = []
        for i, j in fs_filter.items():
            if isinstance(j, RelatedFilter):
                lookup_expression_list.append(i)
                self.get_lookup_expression(j.filterset.get_filters(), related_field=i,
                                           lookup_expression_list=lookup_expression_list)
            else:
                if related_field:
                    lookup_expression_list.append('{}__{}'.format(related_field, i))
                else:
                    lookup_expression_list.append(i)
        return lookup_expression_list

    def dispatch(self, request, *args, **kwargs):
        for field, val in self.request.GET.dict().items():
            if field in getattr(settings, 'INVALID_LOOKUP_SKIP_LIST',  ['format', 'fields']):
                continue
            if self.filter_class:
                # if filter_class is available, return error if any query parameter is not a lookup expression
                valid_fields = self.get_lookup_expression(self.filter_class.get_filters())
                if field not in valid_fields:
                    return JsonResponse(data={'detail': '{} is not a defined filter field in {}'
                                        .format(field, self.model.__name__)}, status=status.HTTP_404_NOT_FOUND)

            elif self.filter_fields:
                # if filter_fields are available, return error if any query parameter is not an available filter field
                if field not in self.filter_fields:
                    return JsonResponse(data={'detail': '{} is not a defined filter field in {}'
                                        .format(field, self.model.__name__)}, status=status.HTTP_404_NOT_FOUND)

            else:
                # if neither filter_class nor filter_fields are available, return error if any query parameter is
                # not a field in the model
                if field.split('__')[0] not in [i.name for i in self.model._meta.fields +
                                                                self.model._meta.many_to_many]:
                    return JsonResponse(data={'detail': '{} is not a valid field in {}'
                                        .format(field, self.model.__name__)}, status=status.HTTP_404_NOT_FOUND)

        return super().dispatch(request, *args, **kwargs)



from django.core.management.base import BaseCommand, CommandError
from django.apps import apps
from django.conf import settings
from jinja2 import Template
import os

__version__ = "0.0.1"


class Command(BaseCommand):
    help = 'Generates files (serializers, views, urls) necessary to host apis via DRF'

    def __init__(self):
        self.app = None
        self.model_list = None
        super(Command, self).__init__()

    def add_arguments(self, parser):
        """ define command arguments """
        parser.add_argument('app', type=str, help='enter the name of the django app')
        parser.add_argument('--api', action='store_true', help='generate views and create apis.py')
        parser.add_argument('--serializer', action='store_true', help='generate serializers and create serializers.py')
        parser.add_argument('--url', action='store_true', help='generate urls and create urls.py')
        parser.add_argument('--output_path', type=str, default=None, help='path where files should be created')
        parser.add_argument('--api_template', type=str, default=None, help='path to Jinja template used to create api')
        parser.add_argument('--serializer_template', type=str, default=None, help='path to Jinja template used to create serializer')
        parser.add_argument('--url_template', type=str, default=None, help='path to Jinja template used to create urls')

    def handle(self, *args, **options):
        """ command entry point """
        if options['app'] not in settings.INSTALLED_APPS:
            raise CommandError('\'{}\' is not an available application in this project'.format(options['app']))

        self.app = options['app']
        self.model_list = self.get_model_list()

        # build serializers file
        if options['serializer']:
            self.build_serializers(output_path=options['output_path'], template_file=options['serializer_template'])

        # build apis file
        if options['api']:
            self.build_apis(output_path=options['output_path'], template_file=options['api_template'])

        # build urls file
        if options['url']:
            self.build_urls(output_path=options['output_path'], template_file=options['url_template'])

        self.stdout.write(self.style.SUCCESS('Files generated!'))

    def get_model_list(self):
        """ return a list of all models in application """
        app = apps.get_app_config(self.app)
        return list(app.get_models())

    @staticmethod
    def get_model_field_names(model, exclude_list=()):
        """ return a list of field names for a given model """
        return [i.name for i in model._meta.fields if type(i).__name__ not in exclude_list]

    def build_serializers(self, output_path=None, template_file=None):
        """ build the serializers.py file for a list of model names """
        if not template_file:
            template_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                         'drf_templates', 'serializers_template.jinja')
        if not output_path:
            output_path = 'serializers.py'
        elif os.path.isdir(output_path):
            output_path = '{}/serializers.py'.format(output_path)

        model_fields = {}
        for model in self.model_list:
            model_fields[model.__name__] = self.get_model_field_names(model)

        data = {'import_models': 'my import statement here',
                'model_list': self.model_list,
                'app_name': self.app,
                'models_file': 'models',
                'model_fields': model_fields,
                'field_list': {'get a list of fields in the model'}
                }
        with open(template_file) as f:
            template = Template(f.read())
        file_text = template.render(data)
        with open(output_path, 'w') as f:
            f.write(file_text)

    def build_apis(self, output_path=None, template_file=None):
        """ build the apis.py (viewsets) file for a list of model names """
        if not template_file:
            template_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                         'drf_templates', 'apis_template.jinja')
        if not output_path:
            output_path = 'apis.py'
        elif os.path.isdir(output_path):
            output_path = '{}/apis.py'.format(output_path)

        model_fields = {}
        for model in self.model_list:
            model_fields[model.__name__] = self.get_model_field_names(model)

        data = {'model_list': self.model_list,
                'app_name': self.app,
                'models_file': 'models',
                'model_fields': model_fields,
                'serializers_file': 'serializers',
                'viewset_type': 'viewsets.ReadOnlyModelViewSet',
                }
        with open(template_file) as f:
            template = Template(f.read())
        file_text = template.render(data)
        with open(output_path, 'w') as f:
            f.write(file_text)

    def build_urls(self, output_path=None, template_file=None):
        """ build the urls.py file for a list of model names """
        if not template_file:
            template_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                         'drf_templates', 'urls_template.jinja')
        if not output_path:
            output_path = 'urls.py'
        elif os.path.isdir(output_path):
            output_path = '{}/urls.py'.format(output_path)

        data = {'import_models': 'my import statement here',
                'model_list': self.model_list,
                'app_name': self.app,
                'views_file': 'apis',
                }
        with open(template_file) as f:
            template = Template(f.read())
        file_text = template.render(data)
        with open(output_path, 'w') as f:
            f.write(file_text)



"""
Description:
    Collection of helpers to assist with permissions on views

How to use:
    Add in your view as a mixin and define groups per method in permission_dict.
    Example:

        class MyView(InAnyGroup, View):
            permission_dict = {'POST': ['superusers'],
                               'GET': ['operators'] }

"""

from django.core.exceptions import PermissionDenied
from django.contrib.auth.views import redirect_to_login
from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME


class MethodGroupPermissionBase(object):
    """ Base class for method group permissions """
    def dispatch(self, request, *args, **kwargs):
        if not self.has_permission(request, *args, **kwargs):
            if settings.LOGIN_URL and REDIRECT_FIELD_NAME:
                return redirect_to_login(request.get_full_path(),
                                         settings.LOGIN_URL,
                                         REDIRECT_FIELD_NAME
                                         )
            else:
                raise PermissionDenied
        return super(MethodGroupPermissionBase, self).dispatch(request, *args, **kwargs)


class InAllGroups(MethodGroupPermissionBase):
    """
    Description:
        Restrict access based on request method and user group; user must be in ALL required groups

    Usage:
        add as mixin to class definition and put the following in your viewset:
            permission_dict = {'POST': ['site_operators', 'site_admins'],
                               'GET': ['site_operators'],
                              }

    """

    def has_permission(self, request, *args, **kwargs):
        if not hasattr(self, 'permission_dict'):
            return False
        permission_dict_mapping = getattr(self, 'permission_dict', {})
        permission_dict = permission_dict_mapping.get(request.method, [])
        if permission_dict is None:
            return False
        return set(permission_dict).issubset([i.name for i in request.user.groups.all()])


class InAnyGroup(MethodGroupPermissionBase):
    """
    Description:
        Restrict access based on request method and user group; user can be in ANY required group

    Usage:
        add as mixin to class definition and put the following in your viewset:
            permission_dict = {'POST': ['site_admins'],
                               'GET': ['site_admins', 'site_operators'],
                              }
    """

    def has_permission(self, request, *args, **kwargs):
        if not hasattr(self, 'permission_dict'):
            return False
        permission_dict_mapping = getattr(self, 'permission_dict', {})
        permission_dict = permission_dict_mapping.get(request.method, [])
        return any(group in [i.name for i in request.user.groups.all()] for group in permission_dict)



from django import template
from django.contrib.auth.models import Group
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter(name='has_group')
def has_group(user, group_name):
    try:
        group = Group.objects.get(name=group_name)
        return group in user.groups.all()
    except:
        return None


@register.filter(name='in')
def inside(value, arg):
    try:
        return value in arg
    except:
        return None


@register.filter
def math_add(value, arg):
    if arg:
        return value + arg


@register.filter
def math_subtract(value, arg):
    if arg:
        return value - arg


@register.filter()
def nbsp(value):
    """ allow blank spaces in value """
    return mark_safe("&nbsp;".join(value.split(' ')))


@register.filter
def intcomma(value):
    """ return thousand-separated formatted representation of value """
    if value:
        return "{:,}".format(value)


@register.filter
def currency_usd(value):
    """ return currency (USD) formatted representation of value """
    if value:
        return "${0:,.2f}".format(value)



import datetime
import csv
import xlwt
from django.http import HttpResponse
from django.views.generic import View
from handyhelpers.mixins.view_mixins import FilterByQueryParamsMixin


class CsvExportView(FilterByQueryParamsMixin, View):
    """ dump a queryset to a csv file """
    queryset = None
    filename = None

    def get(self, request):
        try:
            model = self.queryset.model
            if not self.filename:
                self.filename = "{}.csv".format(model._meta.model_name)
            response = HttpResponse(content_type='text/csv')
            cd = 'attachment; filename="{0}"'.format(self.filename)
            response['Content-Disposition'] = cd
            headers = [field.name for field in model._meta.fields]
            writer = csv.DictWriter(response, fieldnames=headers)
            writer.writeheader()

            queryset = self.filter_by_query_params()
            for row in queryset:
                writer.writerow({column: str(getattr(row, column)) for column in headers})
            return response
        except AttributeError:
            return HttpResponse(content_type='text/csv')


class ExcelExportView(FilterByQueryParamsMixin, View):
    """ dump a queryset to an Excel file """
    queryset = None
    filename = None

    def get(self, request):
        try:
            model = self.queryset.model
            if not self.filename:
                self.filename = "{}.xls".format(model._meta.model_name)
            response = HttpResponse(content_type='application/ms-excel')
            response['Content-Disposition'] = 'attachment; filename="{}"'.format(self.filename)

            wb = xlwt.Workbook(encoding='utf-8')
            ws = wb.add_sheet(model._meta.model_name)

            # Sheet header, first row
            row_num = 0

            font_style = xlwt.XFStyle()
            font_style.font.bold = True

            columns = [field.name for field in model._meta.fields]

            for col_num in range(len(columns)):
                ws.write(row_num, col_num, columns[col_num], font_style)

            # Sheet body, remaining rows
            font_style = xlwt.XFStyle()

            queryset = self.filter_by_query_params()
            for row in queryset.values_list():
                row_num += 1
                for col_num in range(len(row)):
                    if type(row[col_num]) == datetime.datetime:
                        cell_data = str(row[col_num])
                    else:
                        cell_data = row[col_num]
                    ws.write(row_num, col_num, cell_data, font_style)
            wb.save(response)
            return response

        except AttributeError:
            return HttpResponse(content_type='application/ms-excel')



"""
This file contains custom filter lookups as per: https://docs.djangoproject.com/en/2.2/howto/custom-lookups/
"""

from django.db.models import Lookup


class NotEqual(Lookup):
    lookup_name = 'ne'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return '%s <> %s' % (lhs, rhs), params
